(**********************************************************************)
(*                                                                    *)
(*                                                                    *)
(* The OCaml archiver                                                 *)
(*                                                                    *)
(*        written by Gerd Stolpmann                                   *)
(*        Release 0.1                                                 *)
(*                                                                    *)
(*        This is free software, see LICENSE for details.             *)
(*                                                                    *)
(*                                                                    *)
(**********************************************************************)

open Settings;;

(* Some common data types. *)

type version = int list;;
  (* Eg. version 4.2.3 is [4;2;3]. Note that [] represents the development
   * version.
   * There is a canonical ordering defined for versions. This is the 
   * lexicographic ordering with the special case that [] is the biggest
   * version at all. If we say that a version is "smaller than" or "bigger
   * than" another version we refer to this canonical ordering.
   * It is often interestant that a version u is a prefix of another version v
   * because we then assume that v is a "patched" version with respect to u,
   * ie. has the same visible (or at least planned) properties but has
   * corrected bugs or is a better implementation. The development version
   * [] is *not* considered a prefix of any other version. It would make
   * no sense to treat [] as a prefix of some other version because this
   * would mean that the other version is a patch of [].
   * We say that x is "much bigger than" u if it is bigger but u is not a
   * prefix of x.
   *)


type strength = V_exact | V_strict | V_dev | V_sloppy | V_any;;
  (* See module Node for explanation. *)

type location = L_depot | L_ws | L_system | L_local;;


(*** some common auxiliary functions ***)


let string_of_version v =
  let rec print v =
    match v with
      [x]     -> string_of_int x
    | x :: v' -> (string_of_int x) ^ "." ^ print v'
  in

  if v = [] then
    ""
  else
    "-" ^ print v
;;


let path_of_version loc name v =
  match loc with
    L_depot ->
      let s = string_of_version v in
      dir_depot ^ "/" ^ name ^ s
  | L_ws ->
      let s = string_of_version v in
      dir_ws ^ "/" ^ name ^ s
  | L_local ->
      "."                      (* ASSUMES THAT WE ARE IN RIGHT DIR *)
  | L_system ->
      ocaml_installation_dir
;;


let prefix_version v v' =
  (* is v a prefix of v'? *)

  let rec p_version v v' =
    match v with
      [] -> true
    | x :: vv -> begin
      	match v' with
	  [] -> false
      	| x' :: vv' -> x=x' & p_version vv vv'
    end
  in

  match v with
    [] -> v'=[]           (* the development version is NOT a prefix of v' *)
  | _  -> p_version v v'

;;


let is_bigger_version v v' =
  (* is v a bigger version than v' ? *)

  let rec is_b_v v v' =
    match v with
      [] -> false
    | x :: vv -> begin
      	match v' with
	  [] -> true
      	| x' :: vv' ->
	    if x=x' then is_b_v vv vv' else x>x'
    end

  in
  if v = [] then v'<>[] else is_b_v v v'
;;

      
let rec is_much_bigger_version v v' =
  (* v is much bigger than v' 
   * if v is bigger than v'   and   v' is not a prefix of v
   *)
  is_bigger_version v v' &
  not (prefix_version v' v)
;;

      
let bigger_version v v' =
  (* return bigger version *)
  if is_bigger_version v v' then v else v'
;;


(*** Some system functions ***)

let handle_unix_error name f =
  try
    f ()
  with
    Unix.Unix_error (e,_,_) ->
      raise (Sys_error (Unix.error_message e ^ ": " ^ name))
;;


let read_directory name =
  
  let rec read_filenames d =
    try
      let f = Unix.readdir d in
      let rest = read_filenames d in
      if f = "." or f = ".." then
	rest
      else
	f::rest
    with
      End_of_file -> []
  in

  let read_dir () =
    let d = Unix.opendir name in
    try
      let fn = read_filenames d in
      Unix.closedir d;
      fn
    with
      x -> 
	Unix.closedir d;
	raise x
  in
    
  handle_unix_error name read_dir
;;

(*
let open_in' fname =
  try
    open_in fname
  with
    Sys_error s ->
      raise (Sys_error (s ^ ": " ^ fname))
;;
*)

let open_in' = open_in;;


(*** general-purpose functions ***)

let rec delete_all p l =
  match l with
    [] -> []
  | a::l' -> if p a then delete_all p l' else a::(delete_all p l')
;;
