(**********************************************************************)
(*                                                                    *)
(*                                                                    *)
(* The OCaml archiver                                                 *)
(*                                                                    *)
(*        written by Gerd Stolpmann                                   *)
(*        Release 0.1                                                 *)
(*                                                                    *)
(*        This is free software, see LICENSE for details.             *)
(*                                                                    *)
(*                                                                    *)
(**********************************************************************)

(* TODO: OCAML system libraries *)

open Scanmake;;

let ocaml_libs = ["unix"; "num"; "str"; "threads"; "graphics"; "dbm" ];;
let ocaml_dir = "/usr/local/lib/ocaml";;




let findlib_1 name version where =
  if List.mem name ocaml_libs then
    ocaml_dir
  else

  let d = Unix.opendir where in
  let rec getfiles () =
    (* get all files for module "name" *)
    try
      let s = Unix.readdir d in
      if s = name or
	 (String.length name < String.length s &&
	  String.sub s 0 (String.length name + 1) = name ^ "-") then
	s :: getfiles()
      else
	getfiles()
    with
      End_of_file -> []
  in
  let files = getfiles() in
  Unix.closedir d;
  
  if files = [] then
    ""
  else

  (* If version is empty try to get the developer version *)

    if version = [] & List.mem name files then
      where ^ "/" ^ name
    else

    (* Find all files following the version naming scheme *)
      let files_v =
      	List.flatten
	  (List.map
	     (fun file ->
	       try
		 [snd (get_module file)]
	       with
		 Scan_error _ ->
		   []
	     )
	     files
          ) in

      (* Sort files by version *)
      let files_v_sorted = Sort.list (fun a b -> is_bigger_version b a) files_v in

      (* Find appropriate version: *)
      let rec find_version rl =
  	match rl with
	  [v] -> v                (* get what you find *)
      	| v :: v' :: rl' ->
	    if is_bigger_version version v then
	      find_version (v' :: rl')
	    else
	      if prefix_version version v & not (prefix_version version v') then
	      	v
	      else
		if is_much_bigger_version v version then
		  v
		else
	      	  find_version (v' :: rl')
      in

      let v_found = find_version files_v_sorted in

      if not (prefix_version version v_found) then
	prerr_endline ("Version " ^ name ^ string_of_version version ^ 
		       " not found, using " ^ name ^
		       string_of_version v_found ^ " instead.");

      where ^ "/" ^ name ^ string_of_version v_found
;;


let findlib name version where =
  try
    findlib_1 name version where
  with
    Unix.Unix_error (e,_,_) ->
      raise (Sys_error (Unix.error_message e))
;;


let dir_depot = lazy (try Sys.getenv "OCAML_LIB_DEPOT" with Not_found -> "");;
let dir_ws    = lazy (try Sys.getenv "OCAML_LIB_WS"    with Not_found -> "");;

let lookup_lib n v look_in_ws =
  let d_depot = Lazy.force dir_depot in
  let d_ws    = Lazy.force dir_ws    in
  let x, trial =
    if look_in_ws & d_ws <> "" then
      	true, findlib n v d_ws 
      else
	true, ""
    in
    if trial = "" then begin
      if d_depot <> "" then
      	false, findlib n v d_depot
      else
  	look_in_ws, ""
    end
    else x, trial
;;



(* Find libraries. *)

exception Error_lib;;

let find_libraries
      (requirements : (string * version) list)
      (look_in_workspace : bool)
    : (string * (version * bool * string)) list =

  (* returns for the list of requirements [libname1, v1; libname2, v2; ...] the
   * total list of libraries, sorted by the "libA uses libB" relation.
   * This list contains in each element:
   * (libname, (v, found_in_workspace?, directory))
   *
   * If you set "look_in_workspace" also libraries in the workspace are found.
   *)

  (* BUGS:
   * - The returned version is often not the actually used version as given
   *   in the directory
   *)


  let reqs_todo = ref (List.map (fun (n,v) -> (n,(v,look_in_workspace)))
                                requirements) in
  let reqs_done = ref [] in
  let reqs_tree = Hashtbl.create 100 in

  let rec p_requires n n' =
    (* Does n require n', directly or indirectly? *)
    try
      let reqs_n = Hashtbl.find reqs_tree n in
      List.exists
	(fun x -> if x = n' then true else p_requires x n')
	reqs_n
    with
      Not_found -> false
  in

  while !reqs_todo <> [] do
    let (n, (v, look_in_ws)) :: reqs_todo' = !reqs_todo in
    reqs_todo := reqs_todo';
    
    let sub_look_in_ws, place = lookup_lib n v look_in_ws in

    try
      let (found_req : string * (version * bool * string)), 
 	   (new_reqs : (string * (version * bool)) list) =
      	if place = "" then begin
        (* could not find the library *)
	  prerr_endline ("Neither depot nor workspace found.");
          raise Error_lib
	end
      	else
        (* Find out the requirements of this library *)
      	  try
	    let ch = open_in (place ^ "/Ocamlmake") in
	    let il = get_checked_info ch in
	    close_in ch;

	    let r = extract_require il "make" in
	    
	    ((n,(v,sub_look_in_ws,place)),
	     List.map (fun (n1,v1) -> (n1,(v1,sub_look_in_ws))) r)
	  with
	    Scan_error s ->
	      prerr_endline ("Errorneous file: " ^
			     place ^ "/Ocamlmake\n" ^
			     s);
	      raise Error_lib
	  | Sys_error s ->
	      prerr_endline ("OS error: " ^
			     s);
	      raise Error_lib
      in

      (* check on circular references *)
      if List.exists
	  (fun (n',_) -> n=n' or p_requires n' n)
	  new_reqs
      then begin
      	prerr_endline ("Circular reference in requirements");
	raise Error_lib
      end
      else begin

        (* add new_reqs and find inconcistencies in library versions *)

      	let rec add_reqs r =
      	  match r with
	    [] ->
	      ()
      	  | (n',(v',p_ws)) :: r' ->
	      begin
	      (* check if n' already done *)
	      	try
	      	  let (v'',b'',p'') = List.assoc n' !reqs_done in
		  if v' <> v'' then begin
		    prerr_endline ("Version conflict between " ^
				   n' ^ string_of_version v' ^ " and " ^
				   n' ^ string_of_version v'');
		    (* Is this randomly the same library? *)
		    if snd (lookup_lib n' v' p_ws) <> p'' then
		      raise Error_lib
		  end
		  else if b'' & not p_ws then begin
		    prerr_endline ("Depot/Workspace conflict for " ^ 
				   n ^ string_of_version v);
		    raise Error_lib
		  end
		  else if not b'' & p_ws & snd (lookup_lib n v p_ws) <> p''
		  then begin
		    prerr_endline ("Depot/Workspace conflict for " ^ 
				   n ^ string_of_version v);
		    raise Error_lib
		  end
	      	with Not_found ->
		  (* check if n' already included in todo list *)
		  try
		    let (v'',p_ws'') = List.assoc n' !reqs_todo in
		    if v' <> v'' then begin
		      (* I think it makes no sense to check on a
		       * depot/workspace conflict in this case
		       *)
		      let ww = bigger_version v' v'' in
		      reqs_todo := 
			 List.map
			   (fun (nn,(vv,pp)) -> if nn=n 
			                        then (nn,(ww,pp))
			                        else (nn,(vv,pp)))
                           !reqs_todo;
		      prerr_endline ("Version conflict between " ^
				     n' ^ string_of_version v' ^ " and " ^
				     n' ^ string_of_version v'' ^
				     "; using " ^ n' ^ string_of_version ww);
		    end 
		    else if p_ws <> p_ws'' then
		      if snd (lookup_lib n v p_ws) <> snd (lookup_lib n v p_ws'') 
		      then begin
		    	prerr_endline ("Depot/Workspace conflict for " ^ 
				       n ^ string_of_version v);
		    	raise Error_lib
		      end
		  with Not_found ->
		    (* include n' in todo list *)
		    reqs_todo := (n',(v',p_ws)) :: !reqs_todo
	      end;
	      add_reqs r'

	(* end of "let add_reqs ..." *)

	in
	add_reqs new_reqs;
      	Hashtbl.add reqs_tree n (List.map (fun (n',_) -> n') new_reqs);
	reqs_done := found_req :: !reqs_done;
      end
    with
      Error_lib -> 
	prerr_endline ("(library " ^ n ^ " left out.)\n");
        reqs_done := (n,(v,false,"")) :: !reqs_done;
      	Hashtbl.add reqs_tree n []
  done;

  (* We have up to here two results:
   * - reqs_done: the list of bindings "library name -> version, directory"
   * - reqs_tree: a hashtable containing for each library the list of
   *     directly required libraries
   *)

  (* sort the result *)

  (* The library with all dependencies comes first *)

  (* This is a topological sorting. The general form of the algorithm:
   *
   * s = <all nodes>
   * r = []
   * while s is not empty do
   *   src = a source of s
   *   s = s \ src
   *   r = r + src
   * done
   *)


  let rec topo_sort n =
    if n=0 then
      []
    else
      (* find source of s. A source is a node which occurs on the left
       * side of the hashtable reqs_tree only
       *)
      let no_s = ref [] in
      let all  = ref [] in 
  	Hashtbl.iter
	  (fun n nlist ->
	    all  := n :: !all;
	    no_s := nlist @ !no_s
          )
	  reqs_tree;
  	!no_s;

      let rec find_source s =
	match s with
	  [] -> failwith "internal error"
	| s0 :: s' ->
	    if List.mem s0 !no_s then
	      find_source s'
	    else
	      s0

      in
      let source = find_source !all in

      (* remove source from reqs_tree *)
      Hashtbl.remove reqs_tree source;

      (* output source and go forth *)
      let (v,b,p) = List.assoc source !reqs_done in
      
      (source,(v,b,p)) :: topo_sort (n-1)
  in
  topo_sort (List.length !reqs_done)
;;

  
