(**********************************************************************)
(*                                                                    *)
(*                                                                    *)
(* The OCaml archiver                                                 *)
(*                                                                    *)
(*        written by Gerd Stolpmann                                   *)
(*        Release 0.1                                                 *)
(*                                                                    *)
(*        This is free software, see LICENSE for details.             *)
(*                                                                    *)
(*                                                                    *)
(**********************************************************************)

(* A "makemodule" writes parts of the resultant Makefile for
 * certain types of object files and object libraries.
 *
 * For each node a new makemodule is instantiated. It has to
 * generate lines for the follwing Makefile parts:
 *
 * - "sterile": remove all objects, inclusive goal objects.
 *   This method has a parameter containing all objects mentioned
 *   in the Types file having the type of this makemodule.
 *   A "rm -f" line should be produced. The "sterile" method
 *   can add additonal files not normally mentioned in the Types file.
 * - "clean": remove all objects but not goal objects, this means
 *   all intermediate objects. 
 * - "generate": Add rules for making the different generated code
 *   files. Furthermore, return a list of code files that have to
 *   be generated.
 * - "dependencies": Add a line calling a dependency generator.
 *   Furthermore return a file name containing the dependencies.
 *   This file is truncated to zero by Ocamlmake itself before
 *   executing the Makefile for one the parts "sterile", "clean",
 *   "generate" or "dependencies". This file is not touched if
 *   Ocamlmake has only to perform one of the tasks "compile" or
 *   "test".
 * - "compile": There are several sub-parts:
 *   * add rules for making the objects independently
 *   * add linking rules
 *   * return the list of object files to make
 *   * return the list of objects/libraries to link
 * - "test": the test rules and commands are not added by makemodules.
 *
 *
 * Examples of makemodules:
 * - make_ocaml_bytecode
 *     compiles and links bytecode libraries and executables
 * - make_ocaml_native
 *     compiles and links native libraries and executables
 * - make_ocaml_c
 *     compiles and links C files which are intended to be linked
 *     with ocaml objects
 * - make_c_native
 *     compiles and links C files generally
 *
 * ??? THINK ABOUT THAT ???
 * Note that there are certain parts which can be done by more than
 * one makemodule. E.g. the ocamlyacc generation is not specific to
 * make_ocaml_bytecode nor make_ocaml_native. To achieve consistency
 * there should be a common superclass which defines common behaviour.
 * In the Types file the generated code files could be marked either
 * belonging to one of the modules; to avoid problems if one of the
 * modules is not available the generated files get the names of all
 * modules. 
 * 
 * A module is allowed to disable itself. For example, the native
 * compiler is not available for all platforms.
 *)

(* HOW TO OBTAIN <libs>:
 * In link directives you can refer to <libs> which are all required
 * libraries having the correct type. Meant are all goals defined in
 * all (directly or indirectly) required nodes. You have to find out
 * for yourself if the goals have the correct type for your linkage.
 * There are service functions returning the type of a goal.
 * 
 * THE IDEA OF THE "Types" file:
 * We have to distinguish between several types of libraries. There are,
 * for example, bytecode libraries, native liberaries resulting from an
 * ocamlopt and native libraries resulting from a cc run. Generated code
 * is another, completely different format. To make it easier to recognize
 * the type of an object the "Types" file records all files having a type.
 * Just before an "ocamlc", "ocamlopt", "cc" or whatever generates a
 * file is executed the file is appended to "Types" together with its
 * type. (Append *before* because an error can cause a half-written
 * output. ???)
 * This must be done in the Makefile. 
 * 
 * examples:
 *
 * indep/%.cmo: src/%.ml indep/%.cmi
 *         echo $@ make_ocaml_bytecode >>Types
 *         ocamlc -c $<
 *         mv `basename $@` indep
 *
 * gen/%.ml: src/%.mly
 *         echo $@ make_ocaml_code >>Types
 *         ocamlyacc $<
 *         mv `basename $@` gen
 *
 * Better if the "echo" is done after compilation/generation, because otherwise
 * the same file may be appended several times. This can happen if the Makefile
 * stops with an error, and the next invocation of 'make' continues...
 * 
 * Problematic to synchronize "Types" with the real situation.
 * - sterile: remove Types.
 * - clean: remove all files from Types which are deleted
 * ...
 * Since we can delete files by hand, too, we should do a synchronization
 * before each Ocamlmake run, and after the run if required (sterile, clean
 * passes). This is just a stat on each mentioned file to see if it has
 * vanished.
 * After the run we should remove duplicate entries which we'll have from
 * all recompiled files. The last entered type counts.
 *)

class virtual cls (n:Node.t) =
  
  val node = n

  virtual name : unit -> string
      (* Name of this makemodule *)


  virtual enabled : unit -> bool

      (* Note that in the following "commands" need not to preceded by
       * tab characters (as required in the resultant Makefile).
       * The "rules" can both contain dependency information and
       * commands, but now the latter must be preceded by tabs.
       *)

  virtual init_rules : unit -> string list
      (* Rules put in front of all other. You can initialize variables
       * and so on.
       *)
  virtual sterile_cmds : string list -> string list
      (* sterile_cmds <files>: return commands removing at least <files>.
       *   <files> contains relative path names.
       *)
  virtual clean_cmds : string list -> string list -> string list
      (* clean_cmds <files> <goals>: return commands removing all intermediate
       *   files. <files> is the list of all files listed in "Types". 
       *   <goals> is the list of all goals. You have to find out
       *   which of the <files> are <goals> and leave them out.
       *   <files> contains relative path names.
       *   <goals> contains pure names.
       *)
  virtual generate_rules : unit -> string list
      (* return rules for the generation of code from meta code *)
(*  virtual generate_objs : unit -> string list *)
      (* return the list of code files which must be generated.
       * This list contains pure names.
       *)
  virtual dependencies_cmds : unit -> string list
      (* return a command calling a dependency generator *)
  virtual dependencies_obj : unit -> string option
      (* return the file name containing the dependencies *)
  virtual compile_rules : unit -> string list
      (* return rules for independent compilation *)
  virtual link_rules : unit -> string list
      (* return rules for linking anything needed *)
  virtual link_objs : unit -> string list
      (* return object names (as pure names) which are result of linking *)
end

(**********************************************************************)

(* Some auxilary functions: *)

let rules_from_file (name : string) : string list =
  (* Reads Makefile rules from a configuration file named <name>,
   * returns all lines.
   *)
  let f = open_in' (Settings.etc_dir ^ "/" ^ name) in

  let rec read () =
    try 
      read_line f :: read()
    with
      End_of_file -> []
  in

  let lines =
    begin
      try
      	read ()
      with
      	x -> 
      	  close_in f;
      	  raise x
    end
  in
  close_in f;
  lines
;;
  


let gen_sterile_cmds 
    (file_mapping : string -> string list) 
    (files : string list)
    : string list
    =
  (* Generator for the "sterile_cmds" method. Creates a "rm -f" command
   * for all files in the set:
   *  U { file_mapping f | f in files }
   * You can also implement "clean_cmds" with this function if file_mapping
   * filters out all goals.
   *)
  let eff_files = List.flatten (List.map file_mapping files) in
  "rm -f " ^ String.concatenate " " eff_files
;;


let dir_base_name_re = Str.regexp "^([^/]*)/([^/]*)$";;

let gen_clean_filter
    (directories : string list)
    (goals : string list)
    (file : string)
    : string list
    =
  (* if <file> has the form "dirname/basename" and "dirname" is listed in
   * <directories> and "basename" is listed in <goals>, the empty list
   * [] is returned, otherwise [ <file> ] is returned.
   * You can use (gen_clean_filter directories goals) as <file_mapping>
   * in gen_sterile_cmds to recognize all goals in the mentioned directories
   * and leave them out in the "rm" command.
   *)
  if Str.string_match dir_base_name_re file 0 then begin
    let dir  = matched_group 1 file in
    let base = matched_group 2 file in
    if List.mem dir directories & List.mem base goals then
      []
    else
      [file]
  end
  else [file]
;;
