(**********************************************************************)
(*                                                                    *)
(*                                                                    *)
(* The OCaml archiver                                                 *)
(*                                                                    *)
(*        written by Gerd Stolpmann                                   *)
(*        Release 0.1                                                 *)
(*                                                                    *)
(*        This is free software, see LICENSE for details.             *)
(*                                                                    *)
(*                                                                    *)
(**********************************************************************)

(* A node describes 
 * - a library in the depot or in the workspace
 * or
 * - the current program.
 *
 * A node has the following primary attributes. Primary attributes cannot
 * be changed after they have been set:
 *
 * - its name. In the case of a library this is the base name of the
 *   library. In the case of a program this is "."
 * - the required version. (Always [] if the node describes a program.)
 * - how exact the actual version must match the required version.
 *   Possible:
 *   exact:   must exactly match
 *   strict:  The required version is a prefix of the actual version,
 *            and there is no other library for which the required version
 *            is a prefix and has a higher version number than the actual
 *            version
 *   dev:     Use a development version if possible, otherwise like "strict".
 *   sloppy:  Try as in "strict". If that fails use a library with a higher
 *            version than requested
 *   any:     Try as in "sloppy", If that fails use a library with a lower
 *            version than requested
 * - where the library/program is searched. Possible: depot, workspace,
 *   system, local. This is a list of locations.
 * 
 * A node may have the following secondary attributes. These may change.
 *
 * - the actual version
 * - the actual location (depot, workspace, local); perhaps a pathname
 * - a cache for its Ocamlmake contents
 *
 * A node plays a double role: As request the node explains what library
 * should be used, as actual node it contains the chosen binding.
 *
 *)

open Str;;
open Settings;;
open Common;;
open Scanmake;;


type t =
  { (*** primary attributes ***)
    name : string;
    req_version : version;
    req_sob : strength;          (* sob = strength of binding *)
    req_loc : location list; 
    (*** secondary attributes ***)
    mutable act_version : version option;
    mutable act_loc : location option;
    mutable act_path : string option;
    mutable ocamlmake : info list option
  } 
;;


(**********************************************************************)

let create name req_version req_sob req_loc =
  { name = name;
    req_version = req_version;
    req_sob = req_sob;
    req_loc = req_loc;
    act_version = None;
    act_loc = None;
    act_path = None;
    ocamlmake = None
  } 
;;


let name n = n.name;;


let are_equal a b =
  a.name = b.name &
  a.req_version = b.req_version &
  a.req_sob = b.req_sob &
  a.req_loc = b.req_loc
;;


(* The following is only for informational purposes. Do not rely on what
 * these functions return because it may be an approximation. The internal
 * representation will be improved someday and then this interface might
 * no longer be appropriate.
 *)

let requested_version n = n.req_version;;
let requested_strength n = n.req_sob;;
let requested_location n = n.req_loc;;


(**********************************************************************)

(* We maintain a cache of all available versions for a given library name
 * in a given place.
 *)

let version_cache_local () =
  let htab = Hashtbl.create 1 in
  Hashtbl.add htab "." [ [] ];
  htab
;;


let version_cache_system () =
  let d = read_directory Settings.ocaml_installation_dir in
  let re = regexp "^\(.*\)\.cma$" in
  let lst =
    List.flatten
      (List.map
	 (fun name -> 
	   if string_match re name 0 then
	     [matched_group 1 name]
	   else
	     []
	 )
	 d
      ) in
  let tab = Hashtbl.create (2 * List.length lst) in
  List.iter
    (fun name -> Hashtbl.add tab name [ Settings.ocaml_version ])
    lst;
  tab
;;


let version_cache = 
  lazy
  ( [ L_depot,  Hashtbl.create 100;
      L_ws,     Hashtbl.create 100;
      L_system, version_cache_system();
      L_local,  version_cache_local()  ]
    : (location * (string, version list) Hashtbl.t) list 
  )
;;


let get_versions name loc =
  (* get a list of all available versions for library "name" in location
   * "loc". This list is sorted in increasing order. If there is no
   * library "name" an empty list is returned.
   *)

  let get_versions_from_dir () =
    (* read directory *)
    let n = 
      match loc with
	L_depot -> Settings.dir_depot
      |	L_ws    -> Settings.dir_ws
      |	L_local  -> raise Not_found
      |	L_system -> raise Not_found
    in
    let files = read_directory n in
    let versions =
      List.flatten
	(List.map
	   (fun fname ->
	     if fname = name then
	       [ [] ]                        (* dev version *)
	     else
	       if (String.length name < String.length fname &&
		   String.sub fname 0 (String.length name + 1) = name ^ "-")
	       then begin
		 (* an ordinary version *)
		 try
		   [snd (get_module fname)]
		 with
		   Scan_error _ -> []
	       end
	       else []
	   )
           files
        )
    in
    Sort.list 
      (fun a b -> is_bigger_version b a) 
      versions
  in

  
  let htab = List.assoc loc (Lazy.force version_cache) in
  try
    Hashtbl.find htab name
  with
    Not_found ->
      try
      	let l = get_versions_from_dir () in
      	Hashtbl.add htab name l;
      	l
      with
	Not_found ->
	  []
;;


let get_located_versions name loc_list =
  (* get a list of located versions for library "name" possibly located
   * in one of the places mentioned in loc_list.
   * A located version is a pair (loc,v) of a location and a version.
   * If there is no version, raise Not_found.
   * The located list is sorted by versions. If there are two or more
   * locations for a single version the location which comes first in
   * loc_list has the highest priority and is included in the result;
   * the other locations are not.
   *)

  let prefer_location la lb =
    (* Do we prefer location la and drop lb on the basis of loc_list? *)
    let rec prefer ll =
      match ll with
	[]         -> failwith "get_located_versions"
      |	loc :: ll' -> if loc = la then true else if loc = lb then false else
	              prefer ll'
    in
    prefer loc_list
  in

  let rec merge 
            (a : (location * version) list) 
            (b : (location * version) list)
            : (location * version) list =
    (* Merge two sorted lists of located versions *)
    match a with
      [] -> b
    | (la,va) :: a' ->
	match b with
	  [] -> a
	| (lb,vb) :: b' ->
	    if va = vb then begin
	      if prefer_location la lb then
		(la,va) :: merge a' b'
	      else
		(lb,vb) :: merge a' b'
	    end
	    else
	      if is_bigger_version va vb then
		(lb,vb) :: merge a b'
	      else
		(la,va) :: merge a' b
  in

  let rec merge_all 
          (x : (location * version) list)
          (l : (location * version) list list)
           : (location * version) list =
    (* Merge the list of sorted lists of located versions *)
    match l with
      [] -> x
    | a :: l' -> merge_all (merge x a) l'
  in

  let loc_versions_list = List.map 
                            (fun loc -> 
                               (List.map (fun v -> loc,v)
                                         (get_versions name loc)))
                            loc_list in
  let loc_versions = merge_all [] loc_versions_list in

  if loc_versions = [] then
    raise Not_found
  else
    loc_versions
;;


(**********************************************************************)

(* Classify versions as "exact", "strict" and so on. *)

let matches_exactly req_v v = (req_v = v);;

let matches_strictly req_v v = prefix_version req_v v;;

let matches_dev req_v v = (v = []) or prefix_version req_v v;;

let matches_sloppily req_v v = not (is_bigger_version req_v v);;

let matches_any req_v v = true;;


let tolerates n v =
  match n.req_sob with
    V_exact  -> matches_exactly  n.req_version v
  | V_strict -> matches_strictly n.req_version v
  | V_dev    -> matches_dev      n.req_version v
  | V_sloppy -> matches_sloppily n.req_version v
  | V_any    -> matches_any      n.req_version v
;;



(* Order sob: *)

let sob_ordering = 
  [ V_exact, 0;
    V_strict, 1;
    V_dev, 2;
    V_sloppy, 3;
    V_any, 4];;

let is_stricter_than sob1 sob2 =
  List.assoc sob1 sob_ordering < List.assoc sob2 sob_ordering
;;

(**********************************************************************)

(* Unify two nodes. Speaking more exactly, the nodes seen as requests
 * are unified which means that it is checked if the two requests can
 * be fulfilled at the same time, and if this is true, the more specific
 * request is the result of the unification.
 *
 * Raises Not_found if unification is not possible.
 *
 * DETAILS:
 * - If node A has sob V_exact the other node B must tolerate the 
 *   requested version of A. If so, the result is A.
 * - If node A has sob V_strict:
 *   + If B has V_strict, too, and the requested version of A is a
 *     prefix of the requested version of B, return B. If otherwise B
 *     is a prefix of A return A. Otherwise fail.
 *   + If B has V_dev: this is the same, but return always a V_strict
 *     node. We must drop the possibility of choosing a development
 *     version.
 *   + If B has V_sloppy: If A is a prefix of B, return (V_strict,
 *     version of B). (This rule is too hard.) If B is less than A,
 *     return A. Otherwise fail.
 *   + If B has V_any: return A. 
 * - If node A has sob V_dev:
 *   This is the same as if A had sob V_strict with the following exceptions:
 *   + If B has V_dev, too, we do not need to drop the development version
 *   + Of course return always V_dev nodes.
 * - If node A has sob V_sloppy:
 *   + If B has V_sloppy, too: take the bigger version of both
 *   + If B has V_any: return A
 * - If both nodes have V_any:
 *   return the bigger version
 *
 * BUGS AND QUESTIONS:
 * - Problems with the combination V_strict/V_sloppy because we cannot
 *   represent a suitable intersection.
 * - Is the operation associative? Should be.
 * - If both nodes have differently ordered lists of possible locations
 *   the order of the first node counts. (--> avoid this)
 *)  

let unify_nodes_version n1 n2 =
  match n1.req_sob with
    V_exact ->
      if tolerates n2 n1.req_version then 
	(V_exact, n1.req_version)
      else 
	raise Not_found
  | (V_strict | V_dev) ->
      begin
	match n2.req_sob with
	  (V_strict | V_dev) ->
	    if prefix_version n1.req_version n2.req_version then
	      (n1.req_sob, n2.req_version)
	    else if prefix_version n2.req_version n1.req_version then
	      (n1.req_sob, n1.req_version)
	    else
	      raise Not_found
	| V_sloppy ->
	    if prefix_version n1.req_version n2.req_version then
	      (n1.req_sob, n2.req_version)
	    else if is_bigger_version n1.req_version n2.req_version then
	      (n1.req_sob, n1.req_version)
	    else
	      raise Not_found
	| V_any ->
	    (n1.req_sob, n1.req_version)
      end
  | V_sloppy ->
      begin
	match n2.req_sob with
	  V_sloppy ->
	    (V_sloppy, bigger_version n1.req_version n2.req_version)
	| V_any ->
	    (V_sloppy, n1.req_version)
      end
  | V_any ->
      (V_any, bigger_version n1.req_version n2.req_version)
;;


let unify_nodes n1 n2 =
  if n1.name <> n2.name then
    failwith "unify_nodes";

  let (sob,v) =
    if is_stricter_than n1.req_sob n2.req_sob then
      unify_nodes_version n1 n2
    else
      unify_nodes_version n2 n1
  in

  let rec intersect l1 l2 =
    match l1 with
      [] -> []
    | x :: l1' -> if List.mem x l2 then x :: intersect l1' l2 
                                   else      intersect l1' l2
    (* BUG? Cannot represent if l2 has different order than l1 *)
  in

  create n1.name v sob (intersect n1.req_loc n2.req_loc)
;;


(**********************************************************************)

(*
 * get actual version, actual location, actual path, info.
 * if already determined just return them (cache semantics).
 * if not, get_located_versions and retrieve what is requested.
 *)

exception Found of (location * version);;

let determine_version n =
  (* Try to determine the version to which the request is bound.
   * Fills in act_{version,loc,path} if possible.
   *)

  let rec first p l =
    match l with
      []    -> ()
    | x::l' -> if p x then raise (Found x) else first p l'
  in

  let last p l = first p (List.rev l) in

  try
    let lv = get_located_versions n.name n.req_loc in  (* may raise Not_found *)
    if n.req_sob = V_exact then begin
      first (fun (l,v) -> v = n.req_version) lv;
      raise Not_found
    end 
    else begin
      (* If V_dev, try development version first *)
      if n.req_sob = V_dev then
	first (fun (l,v) -> v = []) lv;

      (* Try as for V_strict in any case *)
      last (fun (l,v) -> prefix_version n.req_version v) lv;
      
      (* if V_sloppy or V_any: try a bigger version *)
      if n.req_sob = V_sloppy or n.req_sob = V_any then
	first (fun (l,v) -> is_bigger_version v n.req_version) lv;

      (* if V_any: try next smaller version *)
      if n.req_sob = V_any then
	last (fun (l,v) -> true) lv;

      (* Nothing appropriate. *)
      raise Not_found
    end
  with
    Not_found -> ()
  | Found (loc, v) ->
      n.act_version <- Some v;
      n.act_loc     <- Some loc;
      n.act_path    <- Some (path_of_version loc n.name v)
;;


exception Node_not_found of (string * version);;


let actual_entity f n =
  match f n with
    None   -> 
      determine_version n;
      begin
      	match f n with
	  None   -> raise (Node_not_found (n.name,n.req_version))
	| Some v -> v
      end
  | Some v -> v
;;


(* Return the best matching version and related information, or
 * raise Node_not_found if there isn't any matching version.
 *)

let actual_version  = actual_entity (fun n -> n.act_version);;
let actual_location = actual_entity (fun n -> n.act_loc);;
let actual_path     = actual_entity (fun n -> n.act_path);;

(* actual_info:
 *
 * For L_depot, L_ws and L_local return the contents of Ocamlmake file.
 * For L_system: return a "faked" Ocamlmake file only containing a
 *   "goal" directive.
 *)

let actual_info n =
  match n.ocamlmake with
    None ->
      begin
      	let loc  = actual_location n in
  	let path = actual_path n in
	let omake =
  	  match loc with
	    L_system ->
	      [Section ("make", 
		       [ Goal "dbm";
			 Goal "dynlink";
			 Goal "graphics";
			 Goal "nums";
			 Goal "stdlib";
			 Goal "str";
			 Goal "toplevellib";
			 Goal "unix"
		       ])]
  	  | (L_depot | L_ws | L_local) ->
	      let ch = open_in' (path ^ "/Ocamlmake") in
	      begin
      	      	try
	  	  let i =
		    get_checked_info ch
	  	  in
	    	  close_in ch;
	    	  i
      	      	with
	  	  x ->
		    close_in ch;
		    raise x
	      end
	in
	n.ocamlmake <- Some omake;
	omake
      end
  | Some v -> v
;;

(* requirements:
 * returns list of tuples (name, version, sob, locs)
 * with:
 *   name: name of required node
 *   version: version of required node
 *   sob: what "version" means
 *   locs: list of possible locations.
 *
 * TODO: 
 * - it is currently not possible to specify sob. It is always V_dev.
 * - there is only a hardcoded rule to determine locs:
 *   if actual_location is L_ws --> locs = n.req_loc - L_local
 *   if actual_location is L_depot --> locs = n.req.loc - {L_ws, L_local}
 *   Should be possible to specify alternate rules.
 * - only the requirements of the "make" section are read.
 *   ==> New format for Ocamlmake: Only one "require" clause for both
 *   "make" and "run" contexts.
 *)

let requirements n =
  let delete x l = delete_all (fun a->a=x) l in
  let ocamlmake = actual_info n in
 let reqs = extract_require ocamlmake "make" in
  let locs =
    match actual_location n with
      L_local -> n.req_loc
    | L_ws    -> delete L_local n.req_loc
    | L_depot -> delete L_local (delete L_ws n.req_loc)
    | L_system -> [L_system]
  in
  List.map
    (fun (name, vers) -> (name, vers, V_strict, locs))  (* V_strict: TODO *)
    reqs
;;


(* TODO:
 * node.mli
 *)
 
