(**********************************************************************)
(*                                                                    *)
(*                                                                    *)
(* The OCaml archiver                                                 *)
(*                                                                    *)
(*        written by Gerd Stolpmann                                   *)
(*        Release 0.1                                                 *)
(*                                                                    *)
(*        This is free software, see LICENSE for details.             *)
(*                                                                    *)
(*                                                                    *)
(**********************************************************************)

open Scanmake;;
open Findlib;;


exception User_error of string;;

let user_error s = raise (User_error s);;


let opt_make_depot = ref false;;
let opt_make_ws    = ref true;;

let make_pattern = "/home/gerd/zip0/entwicklung/ocamlarchive/Makefile.pattern";;


let standard_header = "# Makefile automatically generated by ocamlmake. Do not edit by hand!";;


let suffix s =
  try
    let k = String.rindex s '.' in
    String.sub s k (String.length s - k)
  with
    Not_found -> ""
;;


let is_carchive s =
  String.length s >= 6 &
  suffix s = ".a" &
  String.sub s 0 3 = "lib"
;;


let read_ocamlmake where =
  try
    let ch = open_in (where ^ "/Ocamlmake") in
    let x = get_checked_info ch in
    close_in ch;
    x
  with
    Scan_error s -> user_error ("In file " ^ where ^ "/Ocamlmake: " ^ s)
  | Sys_error s  -> user_error ("In file " ^ where ^ "/Ocamlmake: " ^ s)
;;


let ref_obj_file f =
  let sfx = suffix f in
  match sfx with
    (".cmo"|".cma")  -> f
  | (".cmx"|".cmxa"|".o"|".a") -> "$(ARCH)/" ^ f
  | _ -> f
;;


(**********************************************************************)
(* make_single                                                        *)
(**********************************************************************)

let make_single where look_in_ws =
  (* assumes that all required libs have been already made. *)
  
  (* Read the "Ocamlmake" file *)
  let il = read_ocamlmake where in

  (* Check if there is a "Makefile" *)
  if Sys.file_exists (where ^ "/Makefile") then begin
    try
      let ch = open_in (where ^ "/Makefile") in
      try
      	let s = input_line ch in
	if s <> standard_header then
	  user_error ("There is an old " ^ where ^ "/Makefile, please remove it");
	close_in ch
      with
	End_of_file -> ()
    with
      Sys_error s  -> user_error ("In file " ^ where ^ "/Makefile: " ^ s)
  end;

  (* get the required libraries *)
  let req_libs = extract_require il "make" in

  let req_dirs =
    lazy
      (List.map
	 (fun (lib,version) ->
	   let lws, dir = lookup_lib lib version look_in_ws in
	   if dir = "" then
	     user_error ("Cannot find library " ^ 
			 lib ^ string_of_version version);
	   dir
         )
	 req_libs
      )
  in

  (* get the required object files on demand *)
  let req_files = 
    lazy
      (List.flatten
         (List.map
	   (fun dir ->
	     let dir_il = read_ocamlmake dir in
	     List.map
	       (fun file -> dir ^ "/" ^ (ref_obj_file file))
	       (extract_goal dir_il "make")
           )
	   (Lazy.force req_dirs)
         )
      )
  in

  (* get the required ocamlc files on demand *)
  let req_files_ocamlc =
    lazy
      (let rf = Lazy.force req_files in
       List.flatten
	 (List.map
	    (fun f -> 
	      let sfx = suffix f in
	      if sfx = ".cmo" or sfx = ".cma" or sfx = ".o" or
		 (sfx = ".a" & is_carchive f) then
		[f]
	      else
		[]
            )
	    rf
         )
      )
  in

  (* get the required ocamlopt files on demand *)
  let req_files_ocamlopt =
    lazy
      (let rf = Lazy.force req_files in
       List.flatten
	 (List.map
	    (fun f -> 
	      let sfx = suffix f in
	      if sfx = ".cmx" or sfx = ".cmxa" or sfx = ".o" or
		 sfx = ".a" then
		[f]
	      else
		[]
            )
	    rf
         )
      )
  in


  (**************************)
  (* Write the new Makefile *)
  (**************************)

  begin
  try
    let ch = open_out (where ^ "/Makefile") in
    output_string ch (standard_header ^ "\n");

    (************************************)
    (*** Include the Makefile pattern ***)
    (************************************)

    let ch_pat = open_in make_pattern in
        begin
      try
      	while true do
	  let s = input_line ch_pat in
	  output_string ch (s ^ "\n")
      	done
      with
      	End_of_file -> ()
    end;
    close_in ch_pat;

    (***************************)    
    (*** Make the "all" goal ***)
    (***************************)    

    output_string ch (".PHONY: compile\n");
    output_string ch ("compile: ");

    let goals = List.map ref_obj_file (extract_goal il "make") in
    output_string ch (String.concat " " goals ^ "\n");

    (*****************************)
    (*** Make the "link" rules ***)
    (*****************************)

    (* TODO: options currently not supported *)
    let link_rules = extract_link il "make" in
    List.iter
      (fun (goal, objects) ->
	
	(* First write the dependency: *)
	output_string ch ((ref_obj_file goal) ^ " : ");
	List.iter 
	  (fun s -> if s <> "<libs>" & s <> "<toploop>" then 
	              output_string ch (ref_obj_file s ^ " "))
	  objects;
        output_string ch "\n";

	(* Decide which linker to use *)
	let use_ocamlc = ref false in
	let use_ocamlopt = ref false in
	let use_custom = ref false in
	let use_mktop = ref false in

    	List.iter
	  (fun s ->
	    let sfx = suffix s in
	    match sfx with
	      ".cmo"  -> use_ocamlc := true
	    | ".cma"  -> use_ocamlc := true
	    | ".cmx"  -> use_ocamlopt := true
	    | ".cmxa" -> use_ocamlopt := true
	    | ".o"    -> use_custom := true
	    | ".a"    -> use_custom := true
	    | _       -> if s = "<toploop>" then begin
		           use_mktop := true;
		           use_custom := true
	                 end
          )
	  objects;

	(* It is not possible to use both linkers *)

	if !use_ocamlc & !use_ocamlopt then
	  user_error ("Invalid mixture of object files when making " ^ goal);

	(* But we must have something to link *)

	if not !use_ocamlc & not !use_ocamlopt then
	  user_error ("Don't know what to do for " ^ goal);

	if !use_ocamlopt & !use_mktop then
	  user_error ("Can't make top loop with ocamlopt: " ^ goal);

	(* If there is <libs> and ocamlc requested, do further checks
	 * on "custom runtime mode"
	 *)

	if List.mem "<libs>" objects & !use_ocamlc & not !use_custom then begin
	  let rf_ocamlc = Lazy.force req_files_ocamlc in
	  if (List.exists (fun f -> 
	                   let sfx = suffix f in sfx = ".o" or sfx = ".a")
		          rf_ocamlc
             )
	  then
	    use_custom := true
	end;

	(* Which type of linking: *)
	let goal_sfx = suffix goal in
	let make_archive =
	  match goal_sfx with
	    ".cma" -> if !use_ocamlc then true else
	                user_error ("Cannot make bytecode archive from native object files")
	  | ".cmxa"-> if !use_ocamlopt then true else
	                user_error ("Cannot make native archive from bytecode object files")
	  | ".cmo" -> user_error "Incremental linking not possible"
	  | ".cmx" -> user_error "Incremental linking not possible"
	  | ".o"   -> user_error "This type of linking is not supported by ocamlmake"
	  | ".a"   -> user_error "This type of linking is not supported by ocamlmake"
	  | _ -> false
	in

	(* Write down the command *)

	if !use_ocamlc then begin
	  if !use_mktop then
	    output_string ch "\t$(OCAMLMKTOP) "
	  else
	    output_string ch "\t$(OCAMLC) "
	end;
	if !use_ocamlopt then
	  output_string ch "\t$(OCAMLOPT) ";
	if !use_custom then
	  output_string ch "-custom ";
	if make_archive then
	  output_string ch "-a ";
	
	output_string ch ("-o " ^ (ref_obj_file goal) ^ " ");

	List.iter
	  (fun s  ->
	    let sfx = suffix s in
	    match sfx with
	      (".cmo"|".cma"|".cmx"|".cmxa"|".o"|".a") -> 
		output_string ch (ref_obj_file s ^ " ")
	    | _ ->
		if s = "<libs>" then begin
		  let lib_files =
		    if !use_ocamlc then 
		      Lazy.force req_files_ocamlc
		    else
		      Lazy.force req_files_ocamlopt in
		  List.iter
		    (fun s ->
		      output_string ch (s ^ " ")
		    )
		    lib_files
		end
		else
		  if s <> "<toploop>" then
		    user_error ("Bad object file: " ^ s)
          )
	  objects;
	output_string ch "\n";
      )
      link_rules;

    (********************************)
    (*** Include the dependencies ***)
    (********************************)

    (* TODO: Do all ocamllex and ocamlyacc runs before getting the
     * dependencies
     *)

(*
 *    let ch_deps =
 *      Unix.open_process_in 
 *	("cd " ^ where ^ "; ocamldep *.ml *.mli") in
 *    begin
 *      try
 *      	while true do
 *	  let s = input_line ch_deps in
 *	  output_string ch (s ^ "\n")
 *      	done
 *      with
 *      	End_of_file -> ()
 *    end;
 *    begin
 *      match Unix.close_process_in ch_deps with
 *      	Unix.WEXITED n   ->
 *	  if n <> 0 then
 *	    user_error "*** 'ocamldep' exited abnormally"
 *      | Unix.WSIGNALED n ->
 *	  user_error "*** 'ocamldep' killed"
 *      | Unix.WSTOPPED n  ->
 *	  user_error "*** 'ocamldep' stopped"
 *    end;
 *)

    (*******************************************)
    (*** Include <RULES>...</RULES> sections ***)
    (*******************************************)

    (* TODO *)

    (********************************)
    (*** Set AUTOINCLUDE variable ***)
    (********************************)

    output_string ch "AUTOINCLUDE =";
    List.iter
      (fun d ->
	output_string ch (" -I " ^ d)
      )
      (Lazy.force req_dirs);
    output_string ch "\n";


    (********************************)
    (*** Include the dependencies ***)
    (********************************)

    output_string ch "include Makefile.depend\n";

    close_out ch;

    (********************************)
    (* Make Makefile.depend         *)
    (********************************)

    let ch_dp = open_out (where ^ "/Makefile.depend") in
    close_out ch_dp
  with

    (*******************************)
    (*** Some common error cases ***)
    (*******************************)

    Sys_error s  -> user_error ("Cannot generate Makefile: " ^ s)
  | Unix.Unix_error (e,_,_) ->
      user_error ("Cannot generate Makefile: OS error - " ^
		  Unix.error_message e)

  end;

  (****************************)
  (*** execute the Makefile ***)
  (****************************)

  begin
    match Unix.system ("cd " ^ where ^ "; make -f Makefile") with
      Unix.WEXITED n   ->
      	if n <> 0 then
	  user_error "*** 'make' exited abnormally"
    | Unix.WSIGNALED n ->
      	user_error "*** 'make' killed"
    | Unix.WSTOPPED n  ->
      	user_error "*** 'make' stopped"
  end

    (****************)
    (*** Be Happy ***)
    (****************)

;;
