(**********************************************************************)
(*                                                                    *)
(*                                                                    *)
(* The OCaml archiver                                                 *)
(*                                                                    *)
(*        written by Gerd Stolpmann                                   *)
(*        Release 0.1                                                 *)
(*                                                                    *)
(*        This is free software, see LICENSE for details.             *)
(*                                                                    *)
(*                                                                    *)
(**********************************************************************)

(* REPOSITORY:
 * 
 * - contains all nodes
 * - can look up nodes by name
 * - knows about topological structure of node relationship
 * - stores for each node all make_modules.
 * 
 * The repository is created in the first phase of an Ocamlmake run.
 * All required libraries are searched and their topological relationship
 * is found out.
 * 
 * In the second phase nodes are re-made if necessary. This is done by
 * writing a Makefile and executing it. The Makefile is a composition of
 * various parts which come from the make_modules stored together with
 * the nodes. The repository is now the informational basis for this task.
 *)

(* For simplicity there is only one repository. *)

type rep_node_t =
  { n : Node.t
  }
;;


let mk_rep_node n =
  { n = n }
;;
    

module NodeAndName =
  struct
    type t = rep_node_t
    type id_t = string
    let id n = Node.name n.n
  end
;;


module Ntopo =
  Topo.Make(NodeAndName)
;;

(* Ntopo is a module whose signature is Topo.S with
 * type key = string, type el_t = rep_node_t.
 * type t is abstract.
 *)

exception Cannot_unify of (Node.t * Node.t)
;;


let rep = Ntopo.create()
;;


let add init_nodes =
  (* Phase I. "init_nodes" is a list of nodes which are needed anyway.
   * Fills the repository with all nodes that are directly or indirectly
   * required by the "init_nodes".
   *
   * There are different kinds of conflicts:
   * - location conflict: two nodes A,B require a third node C. A and B
   *   would look up C in different locations. If there is a common
   *   subset of possible locations use it.
   * - version conflict: two nodes A,B require a third node C. A and B
   *   require different versions of C. As there is normally a 
   *   not so high "strength of version binding" there may be a version
   *   which can be accepted by both A and B. It is tried to figure out
   *   this version.
   *
   * Solving a conflict is a very expensive operation! Possibly A is
   * already bound to C; then B is detected and the compromise is C'.
   * As C is already loaded it is possible that C itself requires 
   * other nodes which might be loaded, too. Replacing C by C' means
   * to remove all specific requirements of C, so this work has to be
   * done again for C'. What this operation makes really expensive is 
   * that there may be further version conflicts introduced by C'.
   *
   * TODO: better algorithm that avoids this. I think this is possible
   * if we do a BFS and sort each level before adding it.
   *
   * Note: node explicitely mentioned in "init_nodes" are never replaced
   * by other versions.
   *)

  Ntopo.clear rep;

  let init_nodes_names = List.map Node.name init_nodes in

  (* next: list of all nodes which have already been added to "rep",
   * but whose predecessors have not been found out.
   *)
  let next = ref init_nodes   in
  List.iter
    (fun n -> Ntopo.add rep (mk_rep_node n))
    init_nodes;

  while !next <> [] do
    let n :: next' = !next in
    next := next';
    
    (* find out predecessors of n *)
    let n_reqs = Node.requirements n in

    (* loop over all required nodes *)
    List.iter
      (fun (r_name, r_version, r_sob, r_locs) ->
	let r_node = Node.create r_name r_version r_sob r_locs in
	(* if node is new just add it *)
	match
	  try
	    Some ((Ntopo.find rep r_name).n)
	  with
	    Not_found -> None
	with
	  None ->
	    (* node is new *)
	    Ntopo.add rep (mk_rep_node r_node);
	    Ntopo.let_le rep (Node.name r_node) (Node.name n);
	    next := !next @ [r_node]
	| Some r_node' ->
	    (* node already known. check on conflicts *)
	    let r_node_unified = 
	      try
		Node.unify_nodes r_node r_node' 
	      with
		Not_found ->
		  raise (Cannot_unify (r_node, r_node'))
	    in
	    if (Node.are_equal r_node r_node_unified) then begin
	      (* No conflict *)
	      Ntopo.let_le rep (Node.name r_node) (Node.name n);
	    end
	    else begin
	      if List.mem r_name init_nodes_names then 
		raise (Cannot_unify (r_node, r_node'))
	      else
		begin

	          (* Solvable conflict *)
	          (* - remove r_node' from !next, add r_node_unified
	           * - remove r_node' from rep and all nodes only depending on 
	           *   r_node'
	           *)

		  (* THIS CODE CURRENTLY LEADS ALWAYS TO PROGRAM ABORTION.
		   * NOT TESTED AT ALL 
		   *)

		  next := 
		     Common.delete_all
		       (fun a -> Node.name a = r_name)
		       !next
		     @
		     [ r_node_unified];
		  let to_delete =
		    Ntopo.private_property rep r_name in
		  List.iter
		    (fun n -> 
		      if Node.name n.n = r_name then
		    	Ntopo.replace rep r_name (mk_rep_node r_node_unified)
		      else
		    	Ntopo.delete rep (Node.name n.n))
		    to_delete
	    	end
	    end
      )
      n_reqs
  done
;;


(* TODO:
 * - Topo.posessed_elements: kleinere elemente, die ausschliesslich von
 *   einem bestimmten Knoten abhaengig sind; dieser besitzt die elemente
 * - Node.are_equal
 * - delete_all --> common (auch in Node benutzt)
 *)


let lookup nname =
  (Ntopo.find rep nname).n
;;


let all_requirements nname =
  (* Phase II *)
  (* returns all nodes which are directly or indirectly required by node
   * with name "nname".
   *)
  let l = ref [] in
  Ntopo.iter_up_at 
    (fun n ->
      if Node.name n.n <> nname then
      	l := n.n :: !l
    )
    rep
    nname;
  !l
;;
