(**********************************************************************)
(*                                                                    *)
(*                                                                    *)
(* The OCaml archiver                                                 *)
(*                                                                    *)
(*        written by Gerd Stolpmann                                   *)
(*        Release 0.1                                                 *)
(*                                                                    *)
(*        This is free software, see LICENSE for details.             *)
(*                                                                    *)
(*                                                                    *)
(**********************************************************************)

(* Scan "Ocamlmake" files. Currently the following features are
 * supported:
 *
 * - Sections are embraced by <SECTIONNAME> and </SECTIONNAME> as in html.
 *   Note that <SECTIONNAME> and </SECTIONNAME> must stand in separate lines;
 *   if not they are not detected.
 * - Comments within sections must be preceded by an '#'.
 * - Comments between sections may be arbitrary.
 * - You can put all "Ocamlmake" sections in regular html files if
 *   you hide these sections from the browser. Example:
 *   <HTML>
 *   <!-- Here comes the Ocamlmake stuff:
 *     <MAKE>
 *     ...
 *     </MAKE>
 *     <SOURCEFILES>
 *     ...
 *     </SOURCEFILES>
 *   -->
 *   <H1> My new OCAML program </H1>
 *   bla bla bla
 *   </HTML>
 * - You can put all "Ocmalmake" sections in OCAML script files:
 *   #! /usr/local/bin/ocamlscript
 *   <RUN>
 *   require xyz-1.2.1
 *   load <libs>
 *   </RUN>
 *   Your script automatically loads all <libs> which are required.
 *
 * - To control exactly which lines of your files are interpreted as
 *   "Ocamlmake" sections put all these sections in a
 *   <OCAMLMAKE>...</OCAMLMAKE> brace. Interpretation stops after </OCAMLMAKE>.
 *   
 * - Section <MAKE>:
 *
 *   Directive                      Effect
 *   -----------------------------------------------------------------
 *   require module-version         marks "module" as being required.
 *                                  It is checked if "module" is existent
 *                                  and if it is up to date. The libraries
 *                                  and object files of "module" which
 *                                  are "goals" of the "module" can now
 *                                  be referenced as <libs>.
 *   goal file1 file2 ...           These are the files that must be made.
 *   link file = obj1 obj2 ...      Says how to make "file" which is
 *                                  either an executable or a library.
 *                                  "file" is linked by ocamlc or ocamlopt.
 *                                  "obj" may be .cmo, .cma, .cmx, .cmxa,
 *                                  .o or .a file formats. You can use
 *                                  <libs> as symbol for all required
 *                                  libraries that must be linked in.
 *                                  You can use the symbol <toploop> to
 *                                  make a new top loop.
 *      NOTE: There are two sorts of .a archives: files following the
 *      name scheme "lib?*.a" are handled as external C libraries;
 *      other files with suffix ".a" are handled as output from ocamlopt.
 *
 *      NOTE: You can put options for ocamlc and ocamlopt in the list
 *      of object files as: <opt option1 option2 ...>.
 *      You can put options for the native C linker (cc) in the list of
 *      object files as: <ccopt option1 option2 ...>.
 *      You can put options for the linker of the OS (ld) in the list
 *      of object files as: <ldopt option1 option2 ...>.
 *
 *   cclink file = obj1 obj2 ...    Says how to make "file" from .o or .a
 *                                  file formats. "file" is linked by an
 *                                  appropriate "cc" command. If file has the
 *                                  name "lib?*.a" a library is made.
 *   ldlink file = obj1 obj2 ...    Uses "ld" directly.
 *   <RULES>                        This sub section is included in the
 *   ...                            resulting makefile as-is.
 *   </RULES>
 *
 * - Section <RUN>:
 *   require module-version         as above
 *   load obj1 obj2 ...             This is only used in scrips. The objects
 *                                  which may be .cmo or .cma file formats.
 *
 *
 * TODO:
 * - <SOURCEFILES> and <BINARYFILES> sections for distributions
 *   (automatically pack an RPM archive)
 * - Automatically make a toploop suitable for your script
 *)

open Str;;
open String;;
open Common;;

exception Scan_error of string;;

type info =
    Section of (string * info list)
  | Line of string
  | Require of (string * version)
  | Goal of string
  | Link of (string * string list)
  | CCLink of (string * string list)
  | LDLink of (string * string list)
  | Load of string
;;


let scan_error s =
  raise (Scan_error s)
;;


let next_line ch =
  let s = input_line ch in
  prerr_endline s;
  s
;;

let known_sections = [ "ocamlmake"; "make"; "rules"; "run" ];;

let get_info ch : info list =
  (* Reads section structure from channel ch. In the output only
   * "Section" and "Line" are used.
   *)
  let foreign_stuff = ref true in
  let last_section = ref false in
  let sections = ref [] in

  let beg_section_re = regexp "^[ \t]*<\([a-zA-Z0-9_]+\)>[ \t]*$" in
  let end_section_re = regexp "^[ \t]*</\([a-zA-Z0-9_]+\)>[ \t]*$" in
  let comment_re     = regexp "^[ \t]*\(#.*\)?$" in

  let rec get_section (name:string) : info list =
    let s = next_line ch in
    if string_match end_section_re s 0 then begin
      if lowercase(matched_group 1 s)=name then
	[]
      else
	scan_error ("Missing </" ^ name ^ ">")
    end else if string_match comment_re s 0 then
      get_section name
    else if string_match beg_section_re s 0 then begin
      let n = lowercase (matched_group 1 s) in
      if List.mem n known_sections then
	let subsection = get_section n in
	Section (n, subsection) :: get_section name
      else
	scan_error ("Unknown section: <" ^ n ^ ">");
    end
    else
      Line s :: get_section name
  in

  while not !last_section do
    try
      let s = next_line ch in
      let name =
      	if string_match beg_section_re s 0 then
	  let n = lowercase (matched_group 1 s) in
	  begin
	    if  List.mem n known_sections  
	    then
	      n
	    else
	      ""
	  end
	else
	  ""
      in
      if name <> "" then begin
	foreign_stuff := false;
	sections := !sections @ [Section (name, get_section name)];
	foreign_stuff := true;
	if name = "ocamlmake" then
	  last_section := true
      end
    with
      End_of_file ->
      	if !foreign_stuff then 
	  last_section := true
	else
	  raise End_of_file
  done;
  match !sections with
    [ Section ("ocamlmake", x) ] -> x
  | _                            -> !sections
;;

let module_re  = regexp "^\([^-]+\)\(\(-.*\)?\)$";;
let version_re = regexp "^-[0-9]+\(.[0-9]+\)*$";;

let get_module (s:string) : (string * int list) =
  if string_match module_re s 0 then begin
    let mod_name = matched_group 1 s in
    let version  = matched_group 2 s in
    let v =
      if version <> "" then
	if string_match version_re version 0 then
	  let rec get_version i =
	    if i >= length version then
	      []
	    else
	      let k = try index_from version (i+1) '.' 
	      with Not_found -> length version in
	      int_of_string (sub version (i+1) (k-i-1)) :: get_version k
	  in
	  get_version 0
	else
	  scan_error ("Bad module identifier: " ^ s)
      else
	[]
    in
    mod_name, v
  end
  else
    scan_error ("Bad module identifier: " ^ s)
;;


let rec postprocess_info (si:info list) =
  let basic_re      = regexp "^[ \t]*\([^ \t]+\)[ \t]+\(.*[^ \t]\)[ \t]*$" in
  let words_re      = regexp "[ \t]+" in

  let basic s =
    if string_match basic_re s 0 then
      (lowercase (matched_group 1 s), matched_group 2 s)
    else
      scan_error ("Don't know what to do with line:\n" ^ s)
  in

  let postprocess_run (i:info) : info list =
    match i with
      Line l -> 
	let (cmd, args) = basic l in begin
	  match cmd with
	    "require" ->
	      let argsl = split words_re args in
	      List.map (fun arg -> Require (get_module arg)) argsl
	  | "load"    ->
	      let argsl = split words_re args in
	      List.map (fun arg -> Load arg) argsl
	  | s ->
	      scan_error ("Command " ^ s ^ " not permitted here")
	end
    | Section (s, si') -> 
	scan_error ("Section <" ^ s ^ "> not permitted here")
  in

  let postprocess_make (i:info) : info list =
    match i with
      Line l ->
	let (cmd,args) = basic l in begin
	  match cmd with
	    "require" ->
	      let argsl = split words_re args in
	      List.map (fun arg -> Require (get_module arg)) argsl
	  | "goal" ->
	      let argsl = split words_re args in
	      List.map (fun arg -> Goal arg) argsl
	  | "link" ->
	      let argsl = split words_re args in
	      begin
		match argsl with
		  goal :: "=" :: r1 :: r2 -> [Link (goal, r1::r2)]
		| _  ->
		    scan_error ("Bad link directive:\n" ^ l)
	      end
	  | "cclink" ->
	      let argsl = split words_re args in
	      begin
		match argsl with
		  goal :: "=" :: r1 :: r2 -> [CCLink (goal, r1::r2)]
		| _  ->
		    scan_error ("Bad link directive:\n" ^ l)
	      end
	  | "ldlink" ->
	      let argsl = split words_re args in
	      begin
		match argsl with
		  goal :: "=" :: r1 :: r2 -> [LDLink (goal, r1::r2)]
		| _  ->
		    scan_error ("Bad link directive:\n" ^ l)
	      end
	  | s ->
	      scan_error ("Command " ^ s ^ " not permitted here")
	end
    | Section ("rules", si') ->
	if List.for_all 
	    (fun x -> match x with Line _ -> true | _ -> false)
	    si'
	then
	  [Section ("rules", si')]
	else
	  scan_error ("The <rules> section must not have subsections")
    | Section (s, si') -> 
	scan_error ("Section <" ^ s ^ "> not permitted here")
  in	      

  List.flatten
    (List.map
       (fun i ->
	 match i with
	   Section ("ocamlmake",si') -> postprocess_info si'
	 | Section ("run", si')      -> [Section ("run",
						  List.flatten
						    (List.map
						       postprocess_run
						       si'))]
	 | Section ("make", si')     -> [Section ("make",
						  List.flatten
						    (List.map
						       postprocess_make
						       si'))]
	 | Section (x, si')          -> scan_error ("Section <" ^ x ^ "> not permitted here")
	 | Line l                    -> scan_error ("Bad line:\n" ^ l)
       )
       si
   )
;;


let get_checked_info ch =
  postprocess_info (get_info ch)
;;


let extract_from_section (il:info list) (sectname:string) (f:info -> 'a)
    : 'a list =
  List.flatten
    (List.map
       (fun x ->
	 match x with
	   Section (n, il') ->
	     if n = sectname then begin
	       List.flatten
		 (List.map
		    (fun y ->
		      try
			[f y]
		      with
			match_failure -> []
		    )
		    il'
                 )
	     end
	     else
	       []
	 | _ ->
	     []
       )
       il
    )
;;


let extract_require (il:info list) (sectname:string) =
  extract_from_section il sectname (fun (Require x) -> x)
;;


let extract_load il sectname =
  extract_from_section il sectname (fun (Load x) -> x)
;;


let extract_link il sectname =
  extract_from_section il sectname (fun (Link x) -> x)
;;


let extract_cclink il sectname =
  extract_from_section il sectname (fun (CCLink x) -> x)
;;


let extract_ldlink il sectname =
  extract_from_section il sectname (fun (LDLink x) -> x)
;;


let extract_goal il sectname =
  extract_from_section il sectname (fun (Goal x) -> x)
;;


